/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 *
 */
package org.jboss.resource.adapter.jdbc.vendor;

import java.io.Serializable;
import java.lang.reflect.Method;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;

import org.jboss.logging.Logger;
import org.jboss.resource.adapter.jdbc.ValidConnectionChecker;

/**
 * Implements check valid connection sql Requires MySQL driver 3.1.8 or later.
 * This should work on just about any version of the database itself but will
 * only be "fast" on version 3.22.1 and later. Prior to that version it just
 * does "SELECT 1" anyhow.
 * 
 * @author <a href="mailto:adrian@jboss.org">Adrian Brock</a>
 * @author <a href="mailto:acoliver ot jbosss dat org">Andrew C. Oliver</a>
 * @author <a href="mailto:jim.moran@jboss.org">Jim Moran</a>
 * @version $Revision: 1.1.2.2 $
 */
public class MySQLValidConnectionChecker implements ValidConnectionChecker, Serializable {
	private static final long serialVersionUID = -2227528634302168878L;

	private static final Logger log = Logger.getLogger(MySQLValidConnectionChecker.class);

	private Method ping;
	private boolean driverHasPingMethod = false;

	// The timeout (apparently the timeout is ignored?)
	private static Object[] params = new Object[] {};

	public MySQLValidConnectionChecker() {
		try {
			Class mysqlConnection = Thread.currentThread().getContextClassLoader().loadClass("com.mysql.jdbc.Connection");
			ping = mysqlConnection.getMethod("ping", new Class[] {});
			if (ping != null) {
				driverHasPingMethod = true;
			}
		} catch (Exception e) {
			log.warn("Cannot resolve com.mysq.jdbc.Connection.ping method.  Will use 'SELECT 1' instead.", e);
		}
	}

	public SQLException isValidConnection(Connection c) {
		//if there is a ping method then use it, otherwise just use a 'SELECT 1' statement
		if (driverHasPingMethod) {
			try {
				ping.invoke(c, params);
			} catch (Exception e) {
				if (e instanceof SQLException) {
					return (SQLException) e;
				} else {
					log.warn("Unexpected error in ping", e);
					return new SQLException("ping failed: " + e.toString());
				}
			}
			
		} else {
			
			Statement stmt = null;
			try {
				stmt = c.createStatement();
				stmt.executeQuery("SELECT 1");
			} catch (Exception e) {
				if (e instanceof SQLException) {
					return (SQLException) e;
				} else {
					log.warn("Unexpected error in ping (SELECT 1)", e);
					return new SQLException("ping (SELECT 1) failed: " + e.toString());
				}	
			} finally {
				//cleanup the Statment
				try {
					if (stmt != null) stmt.close();
				} catch (SQLException e) {
				}
			}
			
		}
		return null;
	}
}
