/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.ejb3;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import javax.ejb.PostConstruct;
import javax.ejb.PreDestroy;
import javax.ejb.Timeout;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import org.jboss.annotation.ejb.PoolClass;
import org.jboss.aop.AspectManager;
import org.jboss.aop.ClassContainer;
import org.jboss.aop.advice.Interceptor;
import org.jboss.aop.joinpoint.ConstructorInvocation;
import org.jboss.ejb3.callback.SessionCallbackHandler;
import org.jboss.ejb3.entity.InjectedEntityManager;
import org.jboss.ejb3.entity.InjectedEntityManagerFactory;
import org.jboss.ejb3.entity.ManagedEntityManagerFactory;
import org.jboss.ejb3.injection.DependsHandler;
import org.jboss.ejb3.injection.EJBHandler;
import org.jboss.ejb3.injection.Injector;
import org.jboss.ejb3.injection.JndiInjectHandler;
import org.jboss.ejb3.injection.PersistenceContextHandler;
import org.jboss.ejb3.injection.PersistenceUnitHandler;
import org.jboss.ejb3.injection.ResourceHandler;
import org.jboss.ejb3.tx.UserTransactionImpl;
import org.jboss.logging.Logger;
import org.jboss.naming.Util;

/**
 * Comment
 * 
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.32.2.5 $
 */
public abstract class EJBContainer extends ClassContainer implements Container
{

   private static final Logger log = Logger.getLogger(EJBContainer.class);

   protected Pool pool;

   protected String ejbName;

   protected ManagedEntityManagerFactory entityManagerFactory;

   protected ObjectName objectName;

   protected int defaultConstructorIndex;

   protected String beanClassName;

   protected ClassLoader classloader;

   protected Injector[] injectors;

   protected Context enc;

   protected Class beanContextClass;

   protected SessionCallbackHandler callbackHandler;

   protected EjbLinkResolver ejbLinkResolver;

   protected Hashtable initialContextProperties;

   protected HashMap envEntries = new HashMap();

   public EJBContainer(String name, AspectManager manager, ClassLoader cl,
                       String beanClassName, String ejbName, Hashtable ctxProperties)
   {
      super(name, manager);
      this.beanClassName = beanClassName;
      this.classloader = cl;
      this.ejbName = ejbName;
      String on = Ejb3Module.BASE_EJB3_JMX_NAME + ",name=" + ejbName;
      try
      {
         objectName = new ObjectName(on);
      }
      catch (MalformedObjectNameException e)
      {
         throw new RuntimeException(e); // To change body of catch statement
         // use Options | File Templates.
      }
      initialContextProperties = ctxProperties;
   }

   public InitialContext getInitialContext()
   {
      try
      {
         if (initialContextProperties == null)
            return new InitialContext();
         else
            return new InitialContext(initialContextProperties);
      }
      catch (NamingException e)
      {
         throw new RuntimeException(e);
      }
   }

   public void addEnvEntry(String name, String type, String value) throws ClassNotFoundException
   {
      envEntries.put(name, getEnvEntryValue(name, type, value));
   }

   public Context getEnc()
   {
      return enc;
   }

   protected Object getEnvEntryValue(String name, String entryType, String value) throws ClassNotFoundException
   {
      Class type = this.classloader.loadClass(entryType);
      if (type == String.class)
      {
         return value;
      }
      else if (type == Integer.class)
      {
         return new Integer(value);
      }
      else if (type == Long.class)
      {
         return new Long(value);
      }
      else if (type == Double.class)
      {
         return new Double(value);
      }
      else if (type == Float.class)
      {
         return new Float(value);
      }
      else if (type == Byte.class)
      {
         return new Byte(value);
      }
      else if (type == Character.class)
      {
         String input = value;
         if (input == null || input.length() == 0)
         {
            return new Character((char) 0);
         }
         else
         {
            if (input.length() > 1)
               // TODO: Add deployment context
               log.warn("Warning character env-entry is too long: binding="
                        + name + " value=" + input);
            return new Character(input.charAt(0));
         }
      }
      else if (type == Short.class)
      {
         return new Short(value);
      }
      else if (type == Boolean.class)
      {
         return new Boolean(value);
      }
      else
      {
         return value;
      }
   }

   public Hashtable getInitialContextProperties()
   {
      return initialContextProperties;
   }

   public EjbLinkResolver getEjbLinkResolver()
   {
      return ejbLinkResolver;
   }

   public void setEjbLinkResolver(EjbLinkResolver ejbLinkResolver)
   {
      this.ejbLinkResolver = ejbLinkResolver;
   }

   public ObjectName getObjectName()
   {
      return objectName;
   }

   public ManagedEntityManagerFactory getManagedEntityManagerFactory()
   {
      return entityManagerFactory;
   }

   public void setManagedEntityManagerFactory(
           ManagedEntityManagerFactory entityManagerFactory)
   {
      this.entityManagerFactory = entityManagerFactory;
   }

   public String getEjbName()
   {
      return ejbName;
   }

   public String getBeanClassName()
   {
      return beanClassName;
   }

   public Class getBeanClass()
   {
      return clazz;
   }

   public Pool getPool()
   {
      return pool;
   }

   public Object construct()
   {
      Interceptor[] cInterceptors = constructorInterceptors[defaultConstructorIndex];
      if (cInterceptors == null)
      {
         try
         {
            return constructors[defaultConstructorIndex].newInstance();
         }
         catch (InstantiationException e)
         {
            throw new RuntimeException(e);
         }
         catch (IllegalAccessException e)
         {
            throw new RuntimeException(e);
         }
         catch (InvocationTargetException e)
         {
            throw new RuntimeException(e);
         }
      }
      ConstructorInvocation invocation = new ConstructorInvocation(
              cInterceptors);

      invocation.setAdvisor(this);
      invocation.setConstructor(constructors[defaultConstructorIndex]);
      try
      {
         return invocation.invokeNext();
      }
      catch (Throwable throwable)
      {
         throw new RuntimeException(throwable);
      }

   }

   public void start() throws Exception
   {
      for (int i = 0; i < constructors.length; i++)
      {
         if (constructors[i].getParameterTypes().length == 0)
         {
            defaultConstructorIndex = i;
            break;
         }
      }
      initializePool();
      createCallbackHandler();

      createEnvEntries();
   }

   public void create() throws Exception
   {
      clazz = classloader.loadClass(beanClassName);
      initializeClassContainer();
   }

   public void stop() throws Exception
   {
   }

   public void destroy() throws Exception
   {
   }

   protected void createEnvEntries() throws NamingException
   {
      Context ctx = (Context) getInitialContext().lookup("java:");

      Iterator names = envEntries.keySet().iterator();
      while (names.hasNext())
      {
         String name = (String) names.next();
         Object value = envEntries.get(name);

         Util.bind(ctx,
                   "java:comp/env/" + name,
                   value);

         Object o = ctx.lookup("java:comp/env/" + name);
      }
   }

   public void initializePool() throws Exception
   {
      PoolClass poolClass = (PoolClass) resolveAnnotation(PoolClass.class);
      Class poolClazz = poolClass.value();
      int maxSize = poolClass.maxSize();
      long timeout = poolClass.timeout();
      pool = (Pool) poolClazz.newInstance();
      pool.initialize(this, beanContextClass, clazz, maxSize, timeout);

      resolveInjectors();
      pool.setInjectors(injectors);
   }

   public void invokePostConstruct(Object bean)
   {
      callbackHandler.postConstruct(bean);
   }

   public void invokePreDestroy(Object bean)
   {
      callbackHandler.preDestroy(bean);
   }

   public void invokePostActivate(Object bean)
   {
      throw new RuntimeException("PostActivate not implemented for container");
   }

   public void invokePrePassivate(Object bean)
   {
      throw new RuntimeException("PostActivate not implemented for container");
   }

   public void invokeInit(Object bean)
   {

   }

   public void invokeInit(Object bean, Class[] initParameterTypes,
                          Object[] initParameterValues)
   {

   }

   public static final String MANAGED_ENTITY_MANAGER_FACTORY = "ManagedEntityManagerFactory";

   public static final String ENTITY_MANAGER_FACTORY = "EntityManagerFactory";

   protected void resolveInjectors() throws Exception
   {
      Context ctx = getInitialContext();
      enc = ThreadLocalENCFactory.create(ctx);
      ThreadLocalENCFactory.push(enc);
      try
      {
         Thread.currentThread().setContextClassLoader(classloader);
         if (getManagedEntityManagerFactory() != null)
         {
            NonSerializableFactory.rebind(enc, "EntityManager",
                                          new InjectedEntityManager(getManagedEntityManagerFactory(),
                                                                    initialContextProperties));
            NonSerializableFactory.rebind(enc, ENTITY_MANAGER_FACTORY,
                                          new InjectedEntityManagerFactory(
                                                  getManagedEntityManagerFactory()
                                                          .getEntityManagerFactory(),
                                                  initialContextProperties));
            NonSerializableFactory.rebind(enc, MANAGED_ENTITY_MANAGER_FACTORY,
                                          getManagedEntityManagerFactory());
         }
         Util.rebind(enc, "UserTransaction", new UserTransactionImpl());

         List list = new ArrayList();
         list.addAll(JndiInjectHandler.loadInjectors(this));
         list.addAll(ResourceHandler.loadInjectors(this));
         list.addAll(EJBHandler.loadInjectors(this));
         list.addAll(DependsHandler.loadInjectors(this));
         list.addAll(PersistenceContextHandler.loadInjectors(this));
         list.addAll(PersistenceUnitHandler.loadInjectors(this));
         injectors = (Injector[]) list.toArray(new Injector[list.size()]);
      }
      finally
      {
         ThreadLocalENCFactory.pop();
      }
   }

   protected void createCallbackHandler()
   {
      try
      {
         callbackHandler = new SessionCallbackHandler(this,
                                                      getHandledCallbacks());
      }
      catch (Exception e)
      {
         throw new RuntimeException("Error creating callback handler for bean "
                                    + beanClassName, e);
      }
   }

   protected Class[] getHandledCallbacks()
   {
      return new Class[]
              {PostConstruct.class, PreDestroy.class, Timeout.class};
   }
}
