/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.ejb3;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;
import javax.naming.Context;
import javax.naming.NameNotFoundException;
import javax.naming.NamingException;
import org.jboss.annotation.ejb.LocalHome;
import org.jboss.annotation.ejb.RemoteHome;
import org.jboss.aop.AspectManager;
import org.jboss.aop.Dispatcher;
import org.jboss.aop.util.MethodHashing;
import org.jboss.naming.NonSerializableFactory;
import org.jboss.naming.Util;

/**
 * Comment
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.12.2.4 $
 */
public abstract class SessionContainer extends EJBContainer
{
   protected ProxyDeployer proxyDeployer;
   protected Map clusterFamilies = new HashMap();

   public SessionContainer(ClassLoader cl, String beanClassName, String ejbName, AspectManager manager, Hashtable ctxProperties)
   {
      super(Ejb3Module.BASE_EJB3_JMX_NAME + ",name=" + ejbName, manager, cl, beanClassName, ejbName, ctxProperties);
      proxyDeployer = new ProxyDeployer(this);
   }

   public void setDefaultClientBinding(String binding)
   {
      proxyDeployer.setDefaultClientBinding(binding);
   }

   public void start() throws Exception
   {
      super.start();
      Context containers = getContainersContext(getInitialContext());
      Context envCtx = Util.createSubcontext(containers, ejbName);
      NonSerializableFactory.rebind(containers, ejbName, this);

      // So that Remoting layer can reference this container easily.
      Dispatcher.singleton.registerTarget(getObjectName().getCanonicalName(), this);
      proxyDeployer.start();
   }

   public static Context getContainersContext(Context ctx)
           throws NamingException
   {

      try
      {
         return (Context) ctx.lookup("containers");
      }
      catch (NameNotFoundException e)
      {
         return ctx.createSubcontext("containers");
      }
   }

   public Map getClusterFamilies()
   {
      return clusterFamilies;
   }

   public void stop() throws Exception
   {
      getInitialContext().unbind("containers/" + ejbName);
      proxyDeployer.stop();
      Dispatcher.singleton.unregisterTarget(getObjectName().getCanonicalName());
   }
   
   protected void createMethodMap()
   {
      super.createMethodMap();
      try
      {
         RemoteHome home = (RemoteHome) resolveAnnotation(RemoteHome.class);
         if (home != null)
         {
            Method[] declaredMethods = home.value().getMethods();
            for (int i = 0; i < declaredMethods.length; i++)
            {
               long hash = MethodHashing.methodHash(declaredMethods[i]);
               advisedMethods.put(hash, declaredMethods[i]);
            }

            declaredMethods = javax.ejb.EJBObject.class.getMethods();
            for (int i = 0; i < declaredMethods.length; i++)
            {
               long hash = MethodHashing.methodHash(declaredMethods[i]);
               advisedMethods.put(hash, declaredMethods[i]);
            }
         }

         LocalHome localHome = (LocalHome) resolveAnnotation(LocalHome.class);
         if (localHome != null)
         {
            Method[] declaredMethods = localHome.value().getMethods();
            for (int i = 0; i < declaredMethods.length; i++)
            {
               long hash = MethodHashing.methodHash(declaredMethods[i]);
               advisedMethods.put(hash, declaredMethods[i]);
            }

            declaredMethods = javax.ejb.EJBLocalObject.class.getMethods();
            for (int i = 0; i < declaredMethods.length; i++)
            {
               long hash = MethodHashing.methodHash(declaredMethods[i]);
               advisedMethods.put(hash, declaredMethods[i]);
            }
         }
      }
      catch (Exception e)
      {
         throw new RuntimeException(e);
      }
   }
   
   protected boolean isHomeMethod(Method method)
   {
      RemoteHome homeAnnotation = (RemoteHome) resolveAnnotation(RemoteHome.class);
      Class home = null;
      if (homeAnnotation != null)
         home = homeAnnotation.value();
      LocalHome localHomeAnnotation = (LocalHome) resolveAnnotation(LocalHome.class);
      Class localHome = null;
      if (localHomeAnnotation != null)
         localHome = localHomeAnnotation.value();

      if (method.getDeclaringClass().getName().equals("javax.ejb.EJBHome"))
         return true;

      if (method.getDeclaringClass().getName().equals("javax.ejb.EJBLocalHome"))
         return true;

      if (home != null && method.getDeclaringClass().getName().equals(home.getName()))
         return true;

      if (localHome != null && method.getDeclaringClass().getName().equals(localHome.getName()))
         return true;


      return false;
   }
   
   protected boolean isEJBObjectMethod(Method method)
   {
      if (method.getDeclaringClass().getName().equals("javax.ejb.EJBObject"))
         return true;

      if (method.getDeclaringClass().getName().equals("javax.ejb.EJBLocalObject"))
         return true;

      return false;
   }
   
   protected boolean isHandleMethod(Method method)
   {
      if (method.getDeclaringClass().getName().equals("javax.ejb.Handle"))
         return true;

      return false;
   }

}
