/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.ejb3.cache.tree;

import javax.ejb.EJBException;
import javax.ejb.EJBNoSuchObjectException;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import org.jboss.annotation.ejb.cache.tree.CacheConfig;
import org.jboss.aop.Advisor;
import org.jboss.cache.CacheException;
import org.jboss.cache.Fqn;
import org.jboss.cache.TreeCacheMBean;
import org.jboss.ejb3.Container;
import org.jboss.ejb3.Pool;
import org.jboss.ejb3.cache.ClusteredStatefulCache;
import org.jboss.ejb3.stateful.StatefulBeanContext;
import org.jboss.mx.util.MBeanProxy;
import org.jboss.mx.util.MBeanServerLocator;
import org.jboss.util.id.GUID;

/**
 * Comment
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.5.2.3 $
 */
public class StatefulTreeCache implements ClusteredStatefulCache
{
   private Pool pool;
   private PassivationTreeCache cache;
   private String cacheNode;
   private StatefulEvictionPolicy eviction;

   public StatefulBeanContext create()
   {
      StatefulBeanContext ctx = null;
      try
      {
         GUID guid = new GUID();
         ctx = (StatefulBeanContext) pool.get();
         ctx.setId(guid);
         cache.put(cacheNode + "/" + guid.toString(), "bean", ctx);
      }
      catch (Exception e)
      {
         throw new EJBException(e);
      }
      return ctx;
   }

   public StatefulBeanContext create(Class[] initTypes, Object[] initValues)
   {
      StatefulBeanContext ctx = null;
      try
      {
         GUID guid = new GUID();
         ctx = (StatefulBeanContext) pool.get(initTypes, initValues);
         ctx.setId(guid);
         cache.put(cacheNode + "/" + guid.toString(), "bean", ctx);
      }
      catch (Exception e)
      {
         throw new EJBException(e);
      }
      return ctx;
   }

   public StatefulBeanContext get(Object key) throws EJBException
   {
      StatefulBeanContext entry = null;
      try
      {
         Object obj = cache.get(cacheNode + "/" + key, "bean");
         //System.out.println("obj: instanceof: " + obj.getClass().getName());
         entry = (StatefulBeanContext) obj;
      }
      catch (CacheException e)
      {
         throw new RuntimeException(e);
      }
      if (entry == null)
      {
         throw new EJBNoSuchObjectException("Could not find Stateful bean: " + key);
      }
      entry.inUse = true;
      entry.lastUsed = System.currentTimeMillis();
      return entry;
   }

   public void remove(Object key)
   {
      StatefulBeanContext ctx = null;
      try
      {
         cache.remove(cacheNode + "/" + key);
      }
      catch (CacheException e)
      {
         throw new RuntimeException(e);
      }
      if (ctx != null) pool.remove(ctx);

   }

   public void finished(StatefulBeanContext ctx)
   {
      synchronized (ctx)
      {
         ctx.inUse = false;
         ctx.lastUsed = System.currentTimeMillis();
         if (ctx.markedForPassivation)
         {
            try
            {
               Fqn fqn = Fqn.fromString(cacheNode + "/" + ctx.getId());
               eviction.passivate(fqn, ctx);
            }
            catch (Exception e)
            {
               throw new RuntimeException(e);
            }
         }
      }
   }

   public void replicate(StatefulBeanContext ctx)
   {
      try
      {
         cache.put(cacheNode + "/" + ctx.getId(), "bean", ctx);
      }
      catch (CacheException e)
      {
         throw new RuntimeException(e);
      }
   }

   public void initialize(Container container) throws Exception
   {
      Advisor advisor = (Advisor) container;
      this.pool = container.getPool();
      CacheConfig config = (CacheConfig) advisor.resolveAnnotation(CacheConfig.class);
      MBeanServer server = MBeanServerLocator.locateJBoss();
      ObjectName cacheON = new ObjectName(config.name());
      TreeCacheMBean mbean = (TreeCacheMBean) MBeanProxy.get(TreeCacheMBean.class, cacheON, server);
      cache = (PassivationTreeCache) mbean.getInstance();
      cacheNode = "/" + container.getEjbName();
      eviction = (StatefulEvictionPolicy) cache.getEvictionPolicy();
      eviction.createRegion(cacheNode, config.maxSize(), config.idleTimeoutSeconds());
   }

   public void start()
   {

   }

   public void stop()
   {
   }
}
