/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.ejb3.injection;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceUnit;
import javax.persistence.PersistenceUnits;
import org.hibernate.SessionFactory;
import org.jboss.aop.Advisor;
import org.jboss.ejb3.Container;
import org.jboss.ejb3.Ejb3Deployment;
import org.jboss.ejb3.entity.InjectedEntityManager;
import org.jboss.ejb3.entity.InjectedEntityManagerFactory;
import org.jboss.ejb3.entity.InjectedSessionFactory;
import org.jboss.ejb3.entity.ManagedEntityManagerFactory;
import org.jboss.naming.Util;

/**
 * Searches bean class for all @Inject and create Injectors
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.1.2.4 $
 */
public class PersistenceUnitHandler
{
   private static void loadPUsAnnotation(Container container) throws Exception
   {
      PersistenceUnits resources = (PersistenceUnits) ((Advisor) container).resolveAnnotation(PersistenceUnits.class);
      if (resources == null) return;
      for (PersistenceUnit ref : resources.value())
      {
         String encName = ref.name();
         if (encName == null || encName.equals(""))
         {
            throw new RuntimeException("JBoss requires name() for class level @PersistenceUnit");
         }

         EntityManagerFactory factory = getEntityManagerFactory(ref, container);
         bindEntityManagerFactoryToEnc(container, encName, factory);
      }
      PersistenceUnit pu = (PersistenceUnit) ((Advisor) container).resolveAnnotation(PersistenceUnit.class);
      if (pu != null)
      {
         String encName = pu.name();
         if (encName == null || encName.equals(""))
         {
            throw new RuntimeException("JBoss requires name() for class level @PersistenceUnit");
         }

         ManagedEntityManagerFactory factory = getManagedEntityManagerFactory(pu, container);
         bindEntityManagerToEnc(container, factory, encName);
      }
   }

   public static ManagedEntityManagerFactory getManagedEntityManagerFactory(PersistenceUnit ref, Container container)
           throws NamingException
   {
      ManagedEntityManagerFactory factory;
      if (ref.unitName() == null || ref.unitName().equals(""))
      {
         factory = container.getManagedEntityManagerFactory();
      }
      else
      {
         InitialContext ctx = container.getInitialContext();
         factory = (ManagedEntityManagerFactory) ctx.lookup(Ejb3Deployment.MANAGED_ENTITY_FACTORY_CONTEXT + "/" + ref.unitName());
      }
      return factory;
   }


   public static EntityManagerFactory getEntityManagerFactory(PersistenceUnit ref, Container container)
           throws NamingException
   {
      EntityManagerFactory factory;
      if (ref.unitName() == null || ref.unitName().equals(""))
      {
         factory = container.getManagedEntityManagerFactory().getEntityManagerFactory();
         return new InjectedEntityManagerFactory(factory, container.getInitialContextProperties());
      }
      else
      {
         InitialContext ctx = container.getInitialContext();
         factory = (EntityManagerFactory) ctx.lookup(Ejb3Deployment.ACTUAL_ENTITY_MANAGER_FACTORY_CONTEXT + "/" + ref.unitName());
         return new InjectedEntityManagerFactory(factory, Ejb3Deployment.ACTUAL_ENTITY_MANAGER_FACTORY_CONTEXT + "/" +  ref.unitName(), container.getInitialContextProperties());
      }
   }


   public static SessionFactory getSessionFactory(PersistenceUnit ref, Container container)
           throws NamingException
   {
      EntityManagerFactory factory;
      if (ref.unitName() == null || ref.unitName().equals(""))
      {
         factory = container.getManagedEntityManagerFactory().getEntityManagerFactory();
         return new InjectedSessionFactory(factory, container.getInitialContextProperties());
      }
      else
      {
         InitialContext ctx = container.getInitialContext();
         factory = (EntityManagerFactory) ctx.lookup(Ejb3Deployment.ACTUAL_ENTITY_MANAGER_FACTORY_CONTEXT + "/" + ref.unitName());
         return new InjectedSessionFactory(factory, Ejb3Deployment.ACTUAL_ENTITY_MANAGER_FACTORY_CONTEXT + "/" + ref.unitName(), container.getInitialContextProperties());
      }
   }


   public static List loadInjectors(Container container) throws Exception
   {
      Class clazz = container.getBeanClass();
      Method[] methods = clazz.getMethods();
      ArrayList list = new ArrayList();
      Context resourceCtx = Util.createSubcontext(container.getInitialContext(), Container.ENC_CTX_NAME + "/env/resource");
      for (int i = 0; i < methods.length; i++)
      {
         PersistenceUnit ref = (PersistenceUnit) ((Advisor) container).resolveAnnotation(methods[i], PersistenceUnit.class);
         if (ref != null)
         {
            if (!methods[i].getName().startsWith("set")) throw new RuntimeException("@PersistenceContext can only be used with a set method: " + methods[i]);
            if (methods[i].getParameterTypes().length != 1) throw new RuntimeException("@PersistenceContext can only be used with a set method of one parameter: " + methods[i]);
            String encName = ref.name();
            if (encName == null || encName.equals(""))
            {
               encName = methods[i].getName().substring(3);
            }
            if (methods[i].getParameterTypes()[0].getName().equals(EntityManagerFactory.class.getName()))
            {

               EntityManagerFactory factory = getEntityManagerFactory(ref, container);
               list.add(new EntityManagerFactoryMethodInjector(methods[i], container, factory));
               bindEntityManagerFactoryToEnc(container, encName, factory);
            }
            else
            {
               SessionFactory factory = getSessionFactory(ref, container);
               list.add(new EntityManagerFactoryMethodInjector(methods[i], container, factory));
               bindEntityManagerFactoryToEnc(container, encName, factory);

            }
         }
      }
      loadPUsAnnotation(container);
      loadFieldInjectors(clazz, container, list, resourceCtx);
      return list;
   }

   private static void bindEntityManagerFactoryToEnc(Container container, String encName, Object factory)
           throws NamingException
   {
      Context entityCtx = Util.createSubcontext(container.getInitialContext(), Container.ENC_CTX_NAME + "/env");
      entityCtx.bind(encName, factory);
   }

   private static void bindEntityManagerToEnc(Container container, ManagedEntityManagerFactory factory, String encName)
           throws NamingException
   {
      Context entityCtx = Util.createSubcontext(container.getInitialContext(), Container.ENC_CTX_NAME + "/env");
      InjectedEntityManager entityManager = new InjectedEntityManager(factory, container.getInitialContextProperties());
      entityCtx.bind(encName, entityManager);
   }

   private static void loadFieldInjectors(Class clazz, Container container, ArrayList list, Context resourceCtx) throws Exception
   {
      if (clazz == null || clazz.equals(Object.class)) return;
      loadFieldInjectors(clazz.getSuperclass(), container, list, resourceCtx);
      Field[] fields = clazz.getDeclaredFields();
      for (int i = 0; i < fields.length; i++)
      {
         PersistenceUnit ref = (PersistenceUnit) ((Advisor) container).resolveAnnotation(fields[i], PersistenceUnit.class);
         if (ref != null)
         {
            fields[i].setAccessible(true);
            String encName = ref.name();
            if (encName == null || encName.equals(""))
            {
               encName = fields[i].getName();
            }
            if (fields[i].getType().getName().equals(EntityManagerFactory.class.getName()))
            {
               EntityManagerFactory factory = getEntityManagerFactory(ref, container);
               list.add(new EntityManagerFactoryFieldInjector(fields[i], container, factory));
               bindEntityManagerFactoryToEnc(container, encName, factory);
            }
            else
            {
               SessionFactory factory = getSessionFactory(ref, container);
               list.add(new EntityManagerFactoryFieldInjector(fields[i], container, factory));
               bindEntityManagerFactoryToEnc(container, encName, factory);

            }
         }
      }
   }
}
