/*
 * JBoss, the OpenSource EJB server
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.ejb3.service;

import java.lang.reflect.Method;
import java.rmi.MarshalledObject;
import java.util.Hashtable;
import javax.ejb.TimerService;
import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.InvalidAttributeValueException;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.ObjectName;
import javax.management.ReflectionException;
import org.jboss.annotation.ejb.Management;
import org.jboss.annotation.ejb.Service;
import org.jboss.aop.AspectManager;
import org.jboss.aop.MethodInfo;
import org.jboss.aop.advice.Interceptor;
import org.jboss.aop.joinpoint.Invocation;
import org.jboss.aop.joinpoint.InvocationResponse;
import org.jboss.aop.joinpoint.MethodInvocation;
import org.jboss.aop.util.MethodHashing;
import org.jboss.aop.util.PayloadKey;
import org.jboss.aspects.asynch.FutureHolder;
import org.jboss.ejb3.BeanContext;
import org.jboss.ejb3.EJBContainerInvocation;
import org.jboss.ejb3.Ejb3Module;
import org.jboss.ejb3.SessionContainer;
import org.jboss.ejb3.ThreadLocalENCFactory;
import org.jboss.ejb3.asynchronous.AsynchronousInterceptor;
import org.jboss.ejb3.remoting.IsLocalInterceptor;

/**
 * @author <a href="mailto:kabir.khan@jboss.org">Kabir Khan</a>
 * @version $Revision: 1.8.2.4 $
 */
public class ServiceContainer extends SessionContainer
{
   public static final String BASE_EJB3_SERVICE_JMX_NAME = Ejb3Module.BASE_EJB3_JMX_NAME + ",type=service";
   ServiceMBeanDelegate delegate;
   Object singleton;
   boolean injected;

   public ServiceContainer(ClassLoader cl, String beanClassName, String ejbName, AspectManager manager, Hashtable ctxProperties)
   {
      super(cl, beanClassName, ejbName, manager, ctxProperties);
      beanContextClass = ServiceBeanContext.class;
   }

   public Object getSingleton()
   {
      return singleton;
   }

     
   public void start() throws Exception
   {
      super.start();
      singleton = super.construct();
      resolveInjectors();
      registerManagementInterface();
   }

   public void stop() throws Exception
   {
      super.stop();

      if (delegate != null)
      {
         delegate.unregister();
      }
      injected = false;
   }

   public void destroy() throws Exception
   {
      super.destroy();
   }

   public void initializePool() throws Exception
   {
   }

   public TimerService getTimerService()
   {
      throw new RuntimeException("TimerService not implemented for container");
   }

   public void invokePostConstruct(Object bean)
   {
      //Ignore
   }

   public void invokePreDestroy(Object bean)
   {
      //Ignore
   }

   public void invokeInit(Object bean)
   {
      //Ignore
   }

   /**
    * Performs a synchronous local invocation
    */
   public Object localInvoke(Method method, Object[] args) throws Throwable
   {
      return localInvoke(method, args, null);
   }

   /**
    * Performs a synchronous or asynchronous local invocation
    *
    * @param provider If null a synchronous invocation, otherwise an asynchronous
    */
   public Object localInvoke(Method method, Object[] args, FutureHolder provider) throws Throwable
   {
      ClassLoader oldLoader = Thread.currentThread().getContextClassLoader();
      try
      {
         Thread.currentThread().setContextClassLoader(classloader);
         long hash = MethodHashing.calculateHash(method);
         MethodInfo info = (MethodInfo) methodInterceptors.get(hash);
         if (info == null)
         {
            throw new RuntimeException("Could not resolve beanClass method from proxy call: " + method.toString());
         }
         Interceptor[] aspects = info.interceptors;
         EJBContainerInvocation nextInvocation = new EJBContainerInvocation(info, aspects);
         nextInvocation.setAdvisor(this);
         nextInvocation.setArguments(args);

         nextInvocation = populateInvocation(nextInvocation);

         if (provider != null)
         {
            nextInvocation.getMetaData().addMetaData(AsynchronousInterceptor.ASYNCH, AsynchronousInterceptor.INVOKE_ASYNCH, "YES", PayloadKey.AS_IS);
            nextInvocation.getMetaData().addMetaData(AsynchronousInterceptor.ASYNCH, AsynchronousInterceptor.FUTURE_HOLDER, provider, PayloadKey.AS_IS);
         }
         return nextInvocation.invokeNext();
      }
      finally
      {
         Thread.currentThread().setContextClassLoader(oldLoader);
      }
   }

   public InvocationResponse dynamicInvoke(Object target, Invocation invocation) throws Throwable
   {
      ClassLoader oldLoader = Thread.currentThread().getContextClassLoader();
      EJBContainerInvocation newSi = null;
      try
      {
         Thread.currentThread().setContextClassLoader(classloader);
         MethodInvocation si = (MethodInvocation) invocation;
         MethodInfo info = (MethodInfo) methodInterceptors.get(si.getMethodHash());
         if (info == null)
         {
            throw new RuntimeException("Could not resolve beanClass method from proxy call");
         }
         Interceptor[] aspects = info.interceptors;
         newSi = new EJBContainerInvocation(info, aspects);
         newSi.setArguments(si.getArguments());
         newSi.setMetaData(si.getMetaData());
         newSi.setAdvisor(this);

         newSi = populateInvocation(newSi);
         InvocationResponse response = new InvocationResponse(newSi.invokeNext());

         response.setContextInfo(newSi.getResponseContextInfo());

         // marshall return value
         if (invocation.getMetaData().hasTag(IsLocalInterceptor.IS_LOCAL))
         {
            response = (InvocationResponse)new MarshalledObject(response).get();
         }

         return response;
      }
      finally
      {
         Thread.currentThread().setContextClassLoader(oldLoader);
      }
   }

   protected EJBContainerInvocation populateInvocation(EJBContainerInvocation invocation)
   {
      BeanContext ctx = null;
      try
      {
         ctx = (BeanContext) beanContextClass.newInstance();
         ctx.setContainer(this);
         ctx.setInstance(singleton);

         if (!injected)
         {
            injectDependencies(ctx);
         }

         invocation.setTargetObject(singleton);
         invocation.setBeanContext(ctx);
         return invocation;
      }
      catch (InstantiationException e)
      {
         throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
      }
      catch (IllegalAccessException e)
      {
         throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
      }
   }

   protected synchronized void injectDependencies(BeanContext ctx)
   {
      if (!injected)
      {

         if (injectors != null)
         {
            try
            {
               ThreadLocalENCFactory.push(enc);
            for (int i = 0; i < injectors.length; i++)
            {
               injectors[i].inject(ctx);
            }
            }
            finally
            {
               ThreadLocalENCFactory.pop();
            }
         }
         injected = true;
      }
   }

   // Dynamic MBean implementation --------------------------------------------------

   public Object getAttribute(String attribute) throws AttributeNotFoundException,
                                                       MBeanException, ReflectionException
   {
      return delegate.getAttribute(attribute);
   }

   public void setAttribute(Attribute attribute) throws AttributeNotFoundException,
                                                        InvalidAttributeValueException, MBeanException, ReflectionException
   {
      delegate.setAttribute(attribute);
   }

   public AttributeList getAttributes(String[] attributes)
   {
      return delegate.getAttributes(attributes);
   }

   public AttributeList setAttributes(AttributeList attributes)
   {
      return delegate.setAttributes(attributes);
   }

   public Object invoke(String actionName, Object params[], String signature[])
           throws MBeanException, ReflectionException
   {
      return delegate.invoke(actionName, params, signature);
   }


   public MBeanInfo getMBeanInfo()
   {
      return delegate.getMBeanInfo();
   }


   private void registerManagementInterface()
   {
      try
      {
         Class[] interfaces = getBeanClass().getInterfaces();
         Class intf = null;

         for (int i = 0; i < interfaces.length; i++)
         {
            Management ann = (Management) interfaces[i].getAnnotation(Management.class);
            if (ann != null)
            {
               if (intf != null)
               {
                  throw new RuntimeException("Currently only one @Management annotated interface is supported");
               }
               intf = interfaces[i];
            }
         }

         if (intf != null)
         {
            Service service = singleton.getClass().getAnnotation(Service.class);

            ObjectName on = (service.objectName().equals("")) ?
                            new ObjectName(BASE_EJB3_SERVICE_JMX_NAME + ",name=" + ejbName) : new ObjectName(service.objectName());
            delegate = new ServiceMBeanDelegate(this);
            delegate.register(on, intf);
         }
      }
      catch (Exception e)
      {
         throw new RuntimeException("Problem registering @Management interface for @Service " + getBeanClass(), e);
      }
   }

}
