/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.axis.description;

// $Id: ParameterWrapping.java,v 1.1.2.4 2005/07/07 06:38:33 tdiesler Exp $

import java.beans.PropertyDescriptor;
import java.lang.reflect.Method;
import java.util.List;

import javax.xml.namespace.QName;

import org.jboss.axis.enums.Style;
import org.jboss.logging.Logger;

/** A helper class to wrap/unwrap document style request/response structures.
 *
 * @author Thomas.Diesler@jboss.org
 * @since 06-Jun-2005
 */
public class ParameterWrapping
{
   // provide logging
   private static Logger log = Logger.getLogger(ParameterWrapping.class);

   public static Object wrapRequestParameters(OperationDesc opDesc, Object[] inParams)
   {
      assertOperationDesc(opDesc);

      ParameterDesc paramDesc = (ParameterDesc)opDesc.getParameters().get(0);
      QName typeQName = paramDesc.getTypeQName();
      Class reqStructType = paramDesc.getJavaType();
      if (reqStructType == null)
         throw new IllegalArgumentException("Cannot obtain java type for: " + typeQName); 
      
      log.debug("wrapRequestParameters: " + reqStructType.getName());
      try
      {
         List varNames = paramDesc.getWrappedVariables();
         Object reqStruct = reqStructType.newInstance();
         for (int i = 0; i < varNames.size(); i++)
         {
            String varName = (String)varNames.get(i);
            PropertyDescriptor pd = new PropertyDescriptor(varName, reqStructType);
            Method method = pd.getWriteMethod();
            Object value = inParams[i];
            log.debug(" " + method.getName() + ": " + (value != null ? value.getClass().getName() : null));
            method.invoke(reqStruct, new Object[]{value});
         }
         
         return reqStruct;
      }
      catch (RuntimeException rte)
      {
         throw rte;
      }
      catch (Exception e)
      {
         throw new IllegalStateException("Cannot wrap request structure: " + e);
      }
   }

   public static Object[] unwrapRequestParameters(OperationDesc opDesc, Object reqStruct)
   {
      assertOperationDesc(opDesc);

      if (reqStruct == null)
         throw new IllegalArgumentException("Request struct cannot be null");

      ParameterDesc paramDesc = (ParameterDesc)opDesc.getParameters().get(0);
      List varNames = paramDesc.getWrappedVariables();
      Class reqStructType = reqStruct.getClass();
      
      log.debug("unwrapRequestParameters: " + reqStructType.getName());
      Object[] inParams = new Object[varNames.size()];
      try
      {
         for (int i = 0; i < varNames.size(); i++)
         {
            String varName = (String)varNames.get(i);
            PropertyDescriptor pd = new PropertyDescriptor(varName, reqStructType);
            Method method = pd.getReadMethod();
            Object value = method.invoke(reqStruct, new Object[]{});
            log.debug(" " + method.getName() + ": " + (value != null ? value.getClass().getName() : null));
            inParams[i] = value;
         }
      }
      catch (RuntimeException rte)
      {
         throw rte;
      }
      catch (Exception e)
      {
         throw new IllegalArgumentException("Cannot unwrap request structure: " + e);
      }

      return inParams;
   }

   public static Object wrapResponseParameter(OperationDesc opDesc, Object outParam)
   {
      assertOperationDesc(opDesc);

      ParameterDesc paramDesc = opDesc.getReturnParamDesc();
      QName typeQName = paramDesc.getTypeQName();
      Class resStructType = paramDesc.getJavaType();
      if (resStructType == null)
         throw new IllegalArgumentException("Cannot obtain java type for: " + typeQName); 
      
      log.debug("wrapResponseParameter: " + resStructType.getName());
      try
      {
         List varNames = paramDesc.getWrappedVariables();
         String varName = (String)varNames.get(0);
         Object resStruct = resStructType.newInstance();
         PropertyDescriptor pd = new PropertyDescriptor(varName, resStructType);
         Method method = pd.getWriteMethod();
         Object value = outParam;
         log.debug(" " + method.getName() + ": " + (value != null ? value.getClass().getName() : null));
         method.invoke(resStruct, new Object[]{value});
         return resStruct;
      }
      catch (RuntimeException rte)
      {
         throw rte;
      }
      catch (Exception e)
      {
         throw new IllegalStateException("Cannot wrap response structure: " + e);
      }
   }

   public static Object unwrapResponseParameter(OperationDesc opDesc, Object resStruct)
   {
      assertOperationDesc(opDesc);

      if (resStruct == null)
         throw new IllegalArgumentException("Response struct cannot be null");

      ParameterDesc paramDesc = opDesc.getReturnParamDesc();
      List varNames = paramDesc.getWrappedVariables();
      Class reqStructType = resStruct.getClass();

      Object retValue = null;
      log.debug("unwrapResponseParameter: " + reqStructType.getName());
      try
      {
         String varName = (String)varNames.get(0);
         PropertyDescriptor pd = new PropertyDescriptor(varName, reqStructType);
         Method method = pd.getReadMethod();
         Object value = method.invoke(resStruct, new Object[]{});
         log.debug(" " + method.getName() + ": " + (value != null ? value.getClass().getName() : null));
         retValue = value;
      }
      catch (RuntimeException rte)
      {
         throw rte;
      }
      catch (Exception e)
      {
         throw new IllegalArgumentException("Cannot unwrap request structure: " + e);
      }

      return retValue;
   }

   private static void assertOperationDesc(OperationDesc opDesc)
   {
      if (opDesc.getStyle() != Style.DOCUMENT)
         throw new IllegalStateException("Unexpected style: " + opDesc);

      if (opDesc.getParameters().size() != 1)
         throw new IllegalStateException("Unexpected number of input parameters for: " + opDesc);

      if (opDesc.getReturnParamDesc() == null)
         throw new IllegalStateException("Unexpected null return parameter: " + opDesc);

      if (opDesc.isWrapParameters() == false)
         throw new IllegalStateException("Operation meta data not marked for wrapping: " + opDesc);

      ParameterDesc paramDesc = (ParameterDesc)opDesc.getParameters().get(0);
      if (paramDesc.getWrappedVariables() == null)
         throw new IllegalStateException("Parameter meta data does not contain a variable name list: " + paramDesc);

      ParameterDesc retMetaData = opDesc.getReturnParamDesc();
      if (retMetaData.getWrappedVariables() == null)
         throw new IllegalStateException("Return parameter meta data does not contain a variable name list: " + retMetaData);
   }
}
