/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.mx.loading;

import java.net.URL;
import java.io.InputStream;
import java.io.IOException;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipEntry;

import org.jboss.logging.Logger;

/**
 A simple service that can be used preload all classes in the classpath
 of the thread context class loader seen in the start method as the thread
 context class loader. A simple xmbean fragment for deploying the service
 is:
 
    <mbean code="org.jboss.mx.loading.ClassPreloadService"
      name="jboss.mx:service=ClassPreloadService"
      xmbean-dd="">
      <xmbean>
         <operation>
            <name>start</name>
         </operation>
      </xmbean>
   </mbean>

 @author Scott.Stark@jboss.org
 @version $Revision: 1.1 $
 */
public class ClassPreloadService
{
   /**
    Load all classes seen the TCL classpath. This entails a scan of every
    archive in the TCL classpath URLs for .class entries.
    */
   public void start()
   {
      ClassLoader loader = Thread.currentThread().getContextClassLoader();
      URL[] cp = ClassLoaderUtils.getClassLoaderURLs(loader);
      Logger log = Logger.getLogger(ClassPreloadService.class);
      boolean trace = log.isTraceEnabled();
      int loadedClasses = 0;
      int loadErrors = 0;
      for(int n = 0; n < cp.length; n ++)
      {
         URL u = cp[n];
         try
         {
            InputStream is = u.openStream();
            ZipInputStream zis = new ZipInputStream(is);
            ZipEntry ze = zis.getNextEntry();
            while (ze != null)
            {
               String name = ze.getName();
               if (name.endsWith(".class"))
               {
                  int length = name.length();
                  String cname = name.replace('/', '.').substring(0, length - 6);
                  try
                  {
                     Class c = loader.loadClass(cname);
                     loadedClasses ++;
                     if (trace)
                        log.trace("loaded class: " + cname);
                  }
                  catch (Throwable e)
                  {
                     loadErrors ++;
                     if( trace )
                        log.trace("Failed to load class, "+e.getMessage());
                  }
               }
               ze = zis.getNextEntry();
            }
            zis.close();
         }
         catch (IOException ignore)
         {
            // Not a jar
         }
      }
      log.info("Loaded "+loadedClasses+" classes, "+loadErrors+" CNFEs");
   }
}
