/*
 * JBoss, the OpenSource J2EE webOS
 * 
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.test;

import java.util.Arrays;

import junit.framework.AssertionFailedError;
import junit.framework.TestCase;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.FileAppender;
import org.apache.log4j.PatternLayout;
import org.jboss.logging.Logger;
import org.jboss.logging.XLevel;

/**
 * Base Test Case.
 * 
 * @author <a href="adrian@jboss.com">Adrian Brock</a>
 * @version $Revision: 1.6.4.2 $
 */
public class BaseTestCase extends TestCase
{
   // Constants -----------------------------------------------------
   
   // Attributes ----------------------------------------------------

   /** The log */
   protected Logger log;
   
   // Static --------------------------------------------------------

   // Constructors --------------------------------------------------

   /**
    * Create a new base test case
    *
    * @param name the test name
    */
   public BaseTestCase(String name)
   {
      super(name);
   }
   
   // Public --------------------------------------------------------

   /**
    * Get the log for this test
    * 
    * @return the log
    */
   public Logger getLog()
   {
      return log;
   }
   
   // TestCase overrides --------------------------------------------

   // TODO: externalize this   
   protected void setUp() throws Exception
   {
      BasicConfigurator.resetConfiguration();
      PatternLayout layout = new PatternLayout("%r %-5p [%c{1}] %m%n");
      ConsoleAppender appender = new ConsoleAppender(layout);
      BasicConfigurator.configure(appender);
      String file = System.getProperty("org.jboss.test.logfile");
      if (file != null)
      {
         FileAppender fileAppender = new FileAppender(layout, file);
         BasicConfigurator.configure(fileAppender);
      }
      configureLogging();
      log = Logger.getLogger(getClass());
      log("Starting");
   }

   protected void tearDown() throws Exception
   {
      log("Stopping");
   }
   
   // Package protected ---------------------------------------------

   // Protected -----------------------------------------------------

   /**
    * Callback for configuring logging at the start of the test
    */
   protected void configureLogging()
   {
   }

   /**
    * Enable trace for a logging category
    *
    * @param name the logging category
    */
   protected void enableTrace(String name)
   {
      org.apache.log4j.Logger.getLogger(name).setLevel(XLevel.TRACE);
   }

   /**
    * Assert two float values are equal
    *
    * @param one the expected value
    * @param two the actual value
    */
   protected void assertEquals(float one, float two)
   {
      assertEquals(one, two, 0f);
   }

   /**
    * Assert two double values are equal
    *
    * @param one the expected value
    * @param two the actual value
    */
   protected void assertEquals(double one, double two)
   {
      assertEquals(one, two, 0f);
   }

   /**
    * Assert to arrays are equal
    * 
    * @param expected the expected array
    * @param actual the actual array
    */
   protected void assertEquals(Object[] expected, Object[] actual)
   {
      if (Arrays.equals(expected, actual) == false)
         throw new AssertionFailedError("expected: " + Arrays.asList(expected) + " actual: " + Arrays.asList(actual));
   }

   /**
    * Assert to arrays are equal
    * 
    * @param context the context
    * @param expected the expected array
    * @param actual the actual array
    */
   protected void assertEquals(String context, Object[] expected, Object[] actual)
   {
      if (Arrays.equals(expected, actual) == false)
         throw new AssertionFailedError(context + " expected: " + Arrays.asList(expected) + " actual: " + Arrays.asList(actual));
   }

   protected void checkThrowable(Class expected, Throwable throwable)
   {
      if (expected.equals(throwable.getClass()) == false)
      {
         log.error("Unexpected throwable", throwable);
         fail("Unexpected throwable: " + throwable);
      }
   }
   
   // Private -------------------------------------------------------

   /**
    * Log an event with the given context
    *
    * @param context the context
    */   
   private void log(String context)
   {
      log.debug("==== " + context + " " + getName() + " ====");
   }
   
   // Inner classes -------------------------------------------------
}
