/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.test.cluster.test;

import java.util.Hashtable;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import junit.framework.Test;

import org.jboss.test.cluster.ejb.CalledHome;
import org.jboss.test.cluster.ejb.CalledRemote;
import org.jboss.test.JBossClusteredTestCase;

/**
 * HA-JNDI clustering tests.
 *
 * @author Jerry Gauthier
 * @version $Revision: 1.0
 */
public class HAJndiTestCase
      extends JBossClusteredTestCase
      
{
   // JNP URL's
   private static final String NODE0_JNDI = System.getProperty("node0.jndi.url");
   private static final String NODE1_JNDI = System.getProperty("node1.jndi.url");
   private static final String NODE0_HAJNDI = System.getProperty("node0.hajndi.url");
   private static final String NODE1_HAJNDI = System.getProperty("node1.hajndi.url");
   
   // BINDING KEYS and VALUES
   private static final String LOCAL0_KEY = "org.jboss.test.cluster.test.Local0Key";
   private static final String LOCAL0_VALUE = "Local0Value";
   private static final String GLOBAL0_KEY = "org.jboss.test.cluster.test.Global0Key";
   private static final String GLOBAL0_VALUE = "Global0Value";
   private static final String CALLED_HOME_KEY = "cluster.ejb.CalledHome";
	
   public HAJndiTestCase(String name)
   {
      super(name);
   }

   public static Test suite() throws Exception
   {
      Test t1 = JBossClusteredTestCase.getDeploySetup(HAJndiTestCase.class, "cross-server.jar");
      return t1;
   }

   /**
    * Test local bindings using local and HA-JNDI lookups
    *
    * @throws Exception
    */
   public void testLocalBinding()
      throws Exception
   {
      getLog().debug("HAJndiTestCase.testLocalBinding()");
      validateUrls();
      
      // bind to node0 locally
      Context naming = getContext(NODE0_JNDI);
      naming.bind(LOCAL0_KEY, LOCAL0_VALUE);
      closeContext(naming);
      
      // lookup binding locally on Node0 - should succeed
      naming = getContext(NODE0_JNDI);
      String value = (String)lookup(naming, LOCAL0_KEY);
      closeContext(naming);
      assertEquals("lookup local binding on same server", LOCAL0_VALUE, value);
      
      // lookup binding locally on Node1 - should fail
      naming = getContext(NODE1_JNDI);
      value = (String)lookup(naming, LOCAL0_KEY);
      closeContext(naming);
      assertNull("lookup local binding on different server in cluster", value);
      
      // lookup binding using HA-JNDI on Node0 - should succeed
      naming = getContext(NODE0_HAJNDI);
      value = (String)lookup(naming, LOCAL0_KEY);
      closeContext(naming);
      assertEquals("lookup local binding on same server using HA-JNDI", LOCAL0_VALUE, value);
      
      // lookup binding using HA-JNDI on Node1 - should succeed
      naming = getContext(NODE1_HAJNDI);
      value = (String)lookup(naming, LOCAL0_KEY);
      closeContext(naming);
      assertEquals("lookup local binding on different server in cluster using HA-JNDI", LOCAL0_VALUE, value);
   }
   
   /**
    * Test HA-JNDI bindings using local and HA-JNDI lookups
    *
    * @throws Exception
    */
   public void testHAJndiBinding()
      throws Exception
   {
      getLog().debug("HAJndiTestCase.testHAJndiBinding()");
      validateUrls();
      
      // bind to node0 using HA-JNDI
      Context naming = getContext(NODE0_HAJNDI);
      naming.bind(GLOBAL0_KEY, GLOBAL0_VALUE);
      closeContext(naming);
      
      // lookup binding locally on Node0 - should fail
      naming = getContext(NODE0_JNDI);
      String value = (String)lookup(naming, GLOBAL0_KEY);
      closeContext(naming);
      assertNull("lookup HA_JNDI binding on same server using local JNDI", value);
      
      // lookup binding locally on Node1 - should fail
      naming = getContext(NODE1_JNDI);
      value = (String)lookup(naming, GLOBAL0_KEY);
      closeContext(naming);
      assertNull("lookup HA_JNDI binding on different server in cluster using local JNDI", value);
      
      // lookup binding using HA-JNDI on Node0 - should succeed
      naming = getContext(NODE0_HAJNDI);
      value = (String)lookup(naming, GLOBAL0_KEY);
      closeContext(naming);
      assertEquals("lookup HA_JNDI binding on same server using HA-JNDI", GLOBAL0_VALUE, value);
      
      // lookup binding using HA-JNDI on Node1 - should succeed
      naming = getContext(NODE1_HAJNDI);
      value = (String)lookup(naming, GLOBAL0_KEY);
      closeContext(naming);
      assertEquals("lookup HA_JNDI binding on different server in cluster using HA-JNDI", GLOBAL0_VALUE, value);
 
   }

   /**
    * Test EJB Bindings using local and HA-JNDI lookups
    *
    * @throws Exception
    */
   public void testEJBBinding()
      throws Exception
   {
      getLog().debug("HAJndiTestCase.testEJBBinding()");
      validateUrls();
      
      // CalledHome EJB is contained in cross-server.jar and bound during deployment
 
      // lookup binding locally on Node0 - should succeed
      Context naming = getContext(NODE0_JNDI);
      CalledHome home = (CalledHome)lookup(naming, CALLED_HOME_KEY);
      if (home != null)
      { // ensure that EJB is operational  
         CalledRemote remote = home.create();
         remote.remove();
      }
      closeContext(naming);
      assertNotNull("lookup EJB binding on same server using local JNDI", home);
      
      // lookup binding locally on Node1 - should succeed
      naming = getContext(NODE1_JNDI);
      home = (CalledHome)lookup(naming, CALLED_HOME_KEY);
      if (home != null)
      {   // ensure that EJB is operational 
         CalledRemote remote = home.create();
         remote.remove();
      }
      closeContext(naming);
      assertNotNull("lookup EJB binding on different server in cluster using local JNDI", home);
      
      // lookup binding using HA-JNDI on Node0 - should succeed
      naming = getContext(NODE0_HAJNDI);
      home = (CalledHome)lookup(naming, CALLED_HOME_KEY);
      if (home != null)
      {   // ensure that EJB is operational 
         CalledRemote remote = home.create();
         remote.remove();
      }
      closeContext(naming);
      assertNotNull("lookup EJB binding on same server using HA-JNDI", home);
      
      // lookup binding using HA-JNDI on Node1 - should succeed
      naming = getContext(NODE1_HAJNDI);
      home = (CalledHome)lookup(naming, CALLED_HOME_KEY);
      if (home != null)
      {   // ensure that EJB is operational 
         CalledRemote remote = home.create();
         remote.remove();
      }
      closeContext(naming);
      assertNotNull("lookup EJB binding on different server in cluster using HA-JNDI", home);

   }   
  
   private Context getContext(String url)
      throws Exception
   {
      Hashtable env = new Hashtable();        
      env.put(Context.INITIAL_CONTEXT_FACTORY, "org.jnp.interfaces.NamingContextFactory");
		env.put(Context.PROVIDER_URL, url);
		env.put(Context.URL_PKG_PREFIXES, "org.jboss.naming:org.jnp.interfaces");
		
		Context naming = new InitialContext (env);
		return naming;

   }
	
   private void closeContext(Context context)
   {
      try 
      {
         context.close();		   
      }
      catch (NamingException e)
      {
         // no action required
      }
   }
	
   private void bind(Context context, String name, Object value)
      throws Exception
   { 
      context.bind(name, value);
   }
	   
   private Object lookup(Context context, String name)
   {	   
      try
      {
         Object o = context.lookup(name);
         log.info(name + " binding value: " + o);
         return o;
      }
      catch (NamingException e)
      {
         log.info("Name " + name + " not found.");
         return null;
      }	   
   }
   
   private void validateUrls()
      throws Exception      
   {
      if (NODE0_JNDI == null)
         throw new Exception("node0.jndi.url not defined.");
         
      if (NODE1_JNDI == null)
         throw new Exception("node1.jndi.url not defined.");
         
      if (NODE0_HAJNDI == null)
         throw new Exception("node0.hajndi.url not defined.");
         
      if (NODE1_HAJNDI == null)
         throw new Exception("node1.hajndi.url not defined.");

   }

}
