/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.web.tomcat.servlets;

import java.io.IOException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServlet;
import javax.servlet.ServletException;
import javax.servlet.ServletContext;
import javax.servlet.RequestDispatcher;
import javax.servlet.ServletConfig;

import org.jboss.logging.Logger;

/**
 * A generic fowarding servlet that obtains the ServletContext for the
 * forwardContext init-param and then obtain a request dispatcher for the
 * request.getPathInfo and forward the request. This allows a global servlet
 * mapping to redirect requests to a common target. An example would be a
 * global web.xml error-pages with a single shared context for the web
 * pages. This requires that the web context crossContext attribute is set to
 * true.
 * 
 * @author Scott.Stark@jboss.org
 * @version $Revision: 1.1 $
 */
public class ContextForwardServlet extends HttpServlet
{
   private static Logger log = Logger.getLogger(ContextForwardServlet.class);
   /** The name of the context to which requests are forwarded */
   private String forwardContext = "/error-pages";

   public void init(ServletConfig config) throws ServletException
   {
      super.init(config);
      String param = config.getInitParameter("forwardContext");
      if( param != null )
         forwardContext = param;
   }

   /**
    * Lookup the ServletContext associated with the forwardContext init-param
    * and then obtain a request dispatcher for the request.getPathInfo and
    * forward the request. This allows a global servlet mapping to redirect
    * requests to a common target.
    * 
    * @param request
    * @param response
    * @throws ServletException
    * @throws IOException
    */ 
   protected void doGet(HttpServletRequest request, HttpServletResponse response)
      throws ServletException, IOException
   {
      boolean trace = log.isTraceEnabled();
      if( trace )
      {
         log.trace("["+forwardContext+"], PathInfo: "+request.getPathInfo()
            + ", QueryString: "+request.getQueryString()
            + ", ContextPath: "+request.getContextPath()
            + ", HeaderNames: "+request.getHeaderNames()
            + ", isCommitted: "+response.isCommitted()
         );
      }
      String path = request.getPathInfo();
      ServletContext sc = getServletContext().getContext(forwardContext);
      if( sc != null )
      {
         if( trace )
            log.trace("Found ServletContext for: "+forwardContext);
         RequestDispatcher rd = sc.getRequestDispatcher(path);
         if( rd != null )
         {
            if( trace )
               log.trace("Found RequestDispatcher for: "+path);
            rd.forward(request, response);
            return;
         }
      }
      throw new ServletException("No RequestDispatcher for: "+forwardContext+"/"+path);
   }

}
