/*
 * JBoss, the OpenSource WebOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.web.tomcat.tc5.session;

import java.io.IOException;
import javax.servlet.http.HttpSession;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

import org.apache.catalina.*;
import org.apache.catalina.connector.Request;
import org.apache.catalina.connector.Response;
import org.apache.catalina.util.LifecycleSupport;
import org.apache.catalina.valves.ValveBase;
import org.jboss.util.NestedRuntimeException;
import org.jboss.logging.Logger;

/**
 * Web request vavle to specifically handle Tomcat jvmRoute using mod_jk(2)
 * module. We assume that the session is set by cookie only for now, i.e., no
 * support of that from URL. Furthermore, the session id has a format of
 * id.jvmRoute where jvmRoute is used by JK module to determine sticky session
 * during load balancing.
 *
 * @author Ben Wang
 * @version $Revision: 1.1.2.1 $
 */
public class JvmRouteValve extends ValveBase implements Lifecycle
{
   // The info string for this Valve
   private static final String info = "JvmRouteValve/1.0";

   protected static Logger log_ = Logger.getLogger(JvmRouteValve.class);

   // Valve-lifecycle_ helper object
   protected LifecycleSupport support = new LifecycleSupport(this);

   protected AbstractJBossManager manager_;

   /**
    * Create a new Valve.
    *
    */
   public JvmRouteValve(AbstractJBossManager manager)
   {
      super();
      manager_ = manager;
   }

   /**
    * Get information about this Valve.
    */
   public String getInfo()
   {
      return info;
   }

   public void invoke(Request request, Response response) throws IOException, ServletException
   {

      // Need to check it before let it through. This is ok because this valve is inserted only
      // when mod_jk option is configured.
      checkJvmRoute(request, response);

      // let the servlet invokation go through
      getNext().invoke(request, response);
   }

   public void checkJvmRoute(Request req, Response res)
      throws IOException, ServletException
   {
      HttpSession session = req.getSession(false);
      if (session != null)
      {
         String sessionId = session.getId();

         // Obtain JvmRoute
         String jvmRoute = manager_.getJvmRoute();
         if (log_.isDebugEnabled())
         {
            log_.debug("checkJvmRoute(): check if need to re-route based on JvmRoute. Session id: " +
               sessionId + " jvmRoute: " + jvmRoute);
         }

         if (jvmRoute == null)
         {
            throw new RuntimeException("JvmRouteValve.checkJvmRoute(): Tomcat JvmRoute is null. " +
               "Need to assign a value in Tomcat server.xml for load balancing.");
         }

         // Check if incoming session id has JvmRoute appended. If not, append it.
         // TODO. We handle only get session id by cookie only
         if (req.isRequestedSessionIdFromURL())
         {
            // Warning but do nothing
            log_.error("JvmRoutValve.checkJvmRoute(): Can't handle clustering where session id is from URL. Will skip.");
         }
         else
         {
            handleJvmRoute(sessionId, jvmRoute, res);
         }
      }
   }

   protected void handleJvmRoute(String sessionId, String jvmRoute, HttpServletResponse response)
   {
      // get requested jvmRoute.
      // TODO. The current format is assumed to be id.jvmRoute. Can be generalized later.
      String requestedJvmRoute = null;
      int ind = sessionId.indexOf(".");
      if (ind > 0)
      {
         requestedJvmRoute = sessionId.substring(sessionId.indexOf(".") + 1, sessionId.length());
      }

      String sid = sessionId;
      if (requestedJvmRoute == null)
      {
         // If this valve is turned on, we assume we have an appendix of jvmRoute. So this request is new.
         sid = sessionId + "." + jvmRoute;
         manager_.setNewSessionCookie(sid, response);
      }
      else if (requestedJvmRoute.equals(jvmRoute))
      {
         return;  // Nothing more needs to be done.
      }
      else
      {
         // We just have a failover since jvmRoute does not match. We will replace the old one with the new one.
         if (log_.isDebugEnabled())
         {
            log_.debug("handleJvmRoute(): We have detected a failover with differen jvmRoute." +
               " old one: " + requestedJvmRoute + " new one: " + jvmRoute + ". Will reset the session id.");
         }
         int index = sessionId.indexOf(".");
         if (index > 0)
         {
            String base = sessionId.substring(0, sessionId.indexOf("."));
            sid = base + "." + jvmRoute;
         }
         else
         {
            throw new RuntimeException("JvmRouteVavle.handleJvmRoute(): session id doesn't JvmRoute.");
         }

         manager_.setNewSessionCookie(sid, response);
         // Change the sessionId with the new one using local jvmRoute
         ClusteredSession session = null;
         try
         {
            session = (ClusteredSession)manager_.findSession(sessionId);
            // change session id with the new one using local jvmRoute and update cluster if needed.
            if( session != null )
            {
               // Note this will trigger a session remove from the super Tomcat class.
               session.resetIdWithRouteInfo(sid);
               if (log_.isDebugEnabled())
               {
                  log_.debug("handleJvmRoute(): changed catalina session to= [" + sid + "] old one= [" + sessionId + "]");
               }
            }
         }
         catch (IOException e)
         {
            if (log_.isDebugEnabled())
            {
               log_.debug("handleJvmRoute(): manager_.findSession() unable to find session= [" + sessionId + "]", e);
            }
            throw new NestedRuntimeException("JvmRouteValve.handleJvmRoute(): cannot find session [" + sessionId + "]", e);
         }
      }
   }

   // Lifecylce-interface
   public void addLifecycleListener(LifecycleListener listener)
   {
      support.addLifecycleListener(listener);
   }

   public void removeLifecycleListener(LifecycleListener listener)
   {
      support.removeLifecycleListener(listener);
   }

   public LifecycleListener[] findLifecycleListeners()
   {
      return support.findLifecycleListeners();
   }

   public void start() throws LifecycleException
   {
      support.fireLifecycleEvent(START_EVENT, this);
   }

   public void stop() throws LifecycleException
   {
      support.fireLifecycleEvent(STOP_EVENT, this);
   }

}
